#include <idc.idc>
//Microsoft C++ RTTI support for IDA
//Version 2.2 20.02.2004 Igor Skochinsky <skochinsky@mail.ru>

//////////////////////////////////////
// Unknown(long ea, long length)
//////////////////////////////////////
// Mark the ea as unknown for a length
// of length, but don't propagate.
static Unknown( ea, length )
{
  auto i;
//  Message("Unknown(%x,%d)\n",ea, length);
  for(i=0; i < length; i++)
     {
       MakeUnkn(ea+i,0);
     }
}

static ForceQword( x ) {  //Make dword, undefine as needed
 if (!MakeQword( x ))
 {
   Unknown(x,8);
   MakeQword(x);
 }
}

static ForceDword( x ) {  //Make dword, undefine as needed
 if (!MakeDword( x ))
 {
   Unknown(x,4);
   MakeDword(x);
 }
}

static ForceWord( x ) {  //Make word, undefine as needed
 if (!MakeWord( x ))
 {
   Unknown(x,2);
   MakeWord( x );
 }
}

static ForceByte( x ) {  //Make byte, undefine as needed
 if (!MakeByte( x ))
 {
   MakeUnkn(x,0);
   MakeByte( x );
 }
}

static SoftOff ( x ) { //Make offset if !=0
 ForceDword(x);
 if (Dword(x)>0 && Dword(x)<=MaxEA()) OpOff(x,0,0);
}

static GetAsciizStr(x)
{
  auto s,c;
  s = "";
  while (c=Byte(x))
  {
    s = form("%s%c",s,c);
    x = x+1;
  }
  return s;
}

//check if Dword(vtbl-4) points to typeinfo record and extract the type name from it
static GetTypeName(vtbl)
{
  auto x, s, c;
  x = Dword(vtbl-4);
  if ((!x) || (x==BADADDR)) return "";
//  if (Dword(x)||Dword(x+4)||Dword(x+8)) return "";
  x = Dword(x+12);
  if ((!x) || (x==BADADDR)) return "";
  s = "";
  x = x+8;
  while (c=Byte(x))
  {
    s = form("%s%c",s,c);
    x = x+1;
  }
  return s;
}

static DwordCmt(x, cmt)
{
  ForceDword(x);
  MakeComm(x, cmt);
}
static OffCmt(x, cmt)
{
  SoftOff(x);
  MakeComm(x, cmt);
}
static StrCmt(x, cmt)
{
  MakeUnkn(x, 0);
  MakeStr(x, BADADDR);
  MakeName(x, "");
  MakeComm(x, cmt);
}
static DwordArrayCmt(x, n, cmt)
{
  Unknown(x,4*n);
  ForceDword(x);
  MakeArray(x,n);
  MakeComm(x, cmt);
}

static MangleNumber(x)
{
//
// 0 = A@
// X = X-1 (1<=X<=10)
// -X = ?(X-1)
// 12 = 'M@'
// 0x0..0xF = 'A'..'P'

  auto s, sign;
  s=""; sign=0;
  if (x<0)
  {
    sign = 1;
    x = -x;
  }  
  if (x==0)
    return "A@";
  else if (x<=10)
    return form("%s%d",sign?"?":"",x-1);
  else
  {
    while (x>0)
    {
      s = form("%s%c",s,'A'+x%16);
      x = x / 16;
    }
    return sign?"?":""+s+"@";
  }
}
static Parse_BCD(x, indent)
{
  auto indent_str,i,a,s;
  indent_str="";i=0;
  while(i<indent)
  {
    indent_str=indent_str+"    ";
    i++;
  }
/*
struct _s_RTTIBaseClassDescriptor
{
    struct TypeDescriptor* pTypeDescriptor; //type descriptor of the class
    DWORD numContainedBases; //number of nested classes following in the array
    struct PMD where;        //some displacement info
    DWORD attributes;        //usually 0, sometimes 10h
};

struct PMD
{
    int mdisp;  //members displacement ???
    int pdisp;  //
    int vdisp;  //vtable displacement ???
};
*/
  Message(indent_str+"0x%08.8X: RTTIBaseClassDescriptor\n", x);
  Message(indent_str+"    pTypeDescriptor:   %08.8Xh (%s)\n", Dword(x), GetAsciizStr(Dword(x)+8));
  Message(indent_str+"    numContainedBases: %08.8Xh\n", Dword(x+4));
  Message(indent_str+"    PMD:               (%d,%d,%d)\n", Dword(x+8), Dword(x+12), Dword(x+16));
  Message(indent_str+"    attributes:        %08.8Xh\n", Dword(x+20));

  OffCmt(x, "pTypeDescriptor");
  DwordCmt(x+4, "numContainedBases");
  DwordArrayCmt(x+8, 3, "PMD");
  DwordCmt(x+20, "attributes");

  s = Parse_TD(Dword(x), indent+1);
  //??_R1A@?0A@A@B@@8 = B::`RTTI Base Class Descriptor at (0,-1,0,0)'
  MakeName(x,"??_R1"+MangleNumber(Dword(x+8))+MangleNumber(Dword(x+12))+
           MangleNumber(Dword(x+16))+MangleNumber(Dword(x+20))+substr(s,4,-1)+'8');
  return s;
}
static Parse_CHD(x, indent)
{
  auto indent_str,i,a,n,p,s;
  indent_str="";i=0;
  while(i<indent)
  {
    indent_str=indent_str+"    ";
    i++;
  }
  //Message(indent_str+"0x%08.8X: RTTIClassHierarchyDescriptor\n", x);
  /*
  struct _s_RTTIClassHierarchyDescriptor
  {
      DWORD signature;      //always zero?
      DWORD attributes;     //bit 0 = multiple inheritance, bit 1 = virtual inheritance
      DWORD numBaseClasses; //number of classes in pBaseClassArray
      struct _s_RTTIBaseClassArray* pBaseClassArray;
  };
  */
  a = Dword(x+4);
  if (a==1)
    p = "(MI)";
  else if (a==2)
    p = "(VI)";
  else if (a==3)
    p = "(MI VI)";
  else
    p="(SI)";

  Message(indent_str+"    signature:         %08.8Xh\n", Dword(x));
  Message(indent_str+"    attributes:        %08.8Xh %s\n", a, p);
  Message(indent_str+"    numBaseClasses:    %08.8Xh\n", n);
  Message(indent_str+"    pBaseClassArray:   %08.8Xh\n", a);

  DwordCmt(x, "signature");
  DwordCmt(x+4, "attributes");
  DwordCmt(x+8, "numBaseClasses");
  OffCmt(x+12, "pBaseClassArray");

  a=Dword(x+12);
  n=Dword(x+8);
  i=0;
  indent=indent+1;
  while(i<n)
  {
    p = Dword(a);
    //Message(indent_str+"    BaseClass[%02d]:  %08.8Xh\n", i, p);
    OffCmt(a, form("BaseClass[%02d]", i));
    if (i==0)
    {
      s = Parse_BCD(p,indent);
      //??_R2A@@8 = A::`RTTI Base Class Array'
      MakeName(a,"??_R2"+substr(s,4,-1)+'8');
      //??_R3A@@8 = A::`RTTI Class Hierarchy Descriptor'
      MakeName(x,"??_R3"+substr(s,4,-1)+'8');
    }
    else
      Parse_BCD(p,indent);
    if (Dword(p+4)>0) //check numContainedBases
      indent++; //nested classes following
    else
      indent--; //end of nested list
    i=i+1;
    a=a+4;
  }
  return s;
}

static Parse_TD(x, indent)
{
  auto indent_str,i,a;
  indent_str="";i=0;
  while(i<indent)
  {
    indent_str=indent_str+"    ";
    i++;
  }
  //Message(indent_str+"0x%08.8X: TypeDescriptor\n", x);
/*
struct TypeDescriptor
{
    void* pVFTable;  //always pointer to type_info::vftable ?
    void* spare;     //seems to be zero for most classes, and default constructor for exceptions
    char name[0];    //mangled name, starting with .?A (.?AV=classes, .?AU=structs)
};
*/
  a = GetAsciizStr(x+8);
  Message(indent_str+"    pVFTable:          %08.8Xh\n", Dword(x));
  Message(indent_str+"    spare:             %08.8Xh\n", Dword(x+4));
  Message(indent_str+"    name:              '%s'\n", a);

  OffCmt(x, "pVFTable");
  OffCmt(x+4, "spare");
  StrCmt(x+8, "name");

  //??_R0?AVA@@@8 = A `RTTI Type Descriptor'
  MakeName(x,"??_R0"+substr(a,1,-1)+"@8");
  return a;
}


static Parse_COL(x, indent)
{
/*
struct _s_RTTICompleteObjectLocator
{
    DWORD signature; //always zero ?
    DWORD offset;    //offset of this vtable in the class ?
    DWORD cdOffset;  //no idea
    struct TypeDescriptor* pTypeDescriptor; //TypeDescriptor of the class
    struct _s_RTTIClassHierarchyDescriptor* pClassDescriptor; //inheritance hierarchy
};*/
  auto indent_str,i,a;
  indent_str="";i=0;
  while(i<indent)
  {
    indent_str=indent_str+"    ";
    i++;
  }
  //Message(indent_str+"0x%08.8X: RTTICompleteObjectLocator\n", x);
  Message(indent_str+"    signature:         %08.8Xh\n", Dword(x));
  Message(indent_str+"    offset:            %08.8Xh\n", Dword(x+4));
  Message(indent_str+"    cdOffset:          %08.8Xh\n", Dword(x+8));
  Message(indent_str+"    pTypeDescriptor:   %08.8Xh (%s)\n", Dword(x+12), GetAsciizStr(Dword(x+12)+8));
  Message(indent_str+"    pClassDescriptor:  %08.8Xh\n", Dword(x+16));

  DwordCmt(x, "signature");
  DwordCmt(x+4, "offset");
  DwordCmt(x+8, "cdOffset");
  OffCmt(x+12, "pTypeDescriptor");
  OffCmt(x+16, "pClassDescriptor");

  //Parse_TD(Dword(x+12),indent+1);
  Parse_CHD(Dword(x+16),indent+1);
}
//get class name for this vtable instance
static GetVtableClass(x)
{
  auto offset, n, i, s, a, p;
  offset = Dword(x+4);
  x = Dword(x+16); //Class Hierarchy Descriptor

  a=Dword(x+12); //pBaseClassArray
  n=Dword(x+8);  //numBaseClasses
  i=0;
  while(i<n)
  {
    p = Dword(a);
    //Message(indent_str+"    BaseClass[%02d]:  %08.8Xh\n", i, p);
    if (Dword(p+8)==offset)
    {
      //found it
      s = GetAsciizStr(Dword(p)+8);
      return s;
    }
    i=i+1;
    a=a+4;
  }
  return s;
}
static ParseVtbl()
{
  auto a,i,s,s2;
  a=ScreenEA();
  s=GetTypeName(a);
  if (substr(s,0,3)==".?A")
  {
    Message("RTTICompleteObjectLocator:      %08.8Xh\n", Dword(a-4));
    Parse_COL(Dword(a-4),0);
    Unknown(a-4,4);
    SoftOff(a-4);
    i = Dword(a-4);  //COL
    i = Dword(i+16); //CHD
    i = Dword(i+4);  //Attributes
    if (i==0)
    { //Single inheritance, so we don't need to worry about duplicate names (several vtables)
      s=substr(s,4,-1);
      MakeName(a,"??_7"+s+"6B@");
      MakeName(Dword(a-4),"??_R4"+s+"6B@");
    }
    else if (i==1)
    { 
      Message("Multiple inheritance\n");
      s2 = GetVtableClass(Dword(a-4));
      s2 = substr(s2,4,-1);
      s  = substr(s,4,-1);
      s = s+"6B"+s2+"@";
      MakeName(a,"??_7"+s);
      MakeName(Dword(a-4),"??_R4"+s);
    }
  }
}

static main(void)
{
 AddHotkey("Alt-F8","ParseVtbl");
 Message("Use Alt-F8 to parse vtable\n");
}
